
{******************************************}
{                                          }
{             FastReport VCL               }
{            PDF export filter             }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}
{      haBlock alignment improved by:      }
{              Nikolay Zverev              }
{            www.delphinotes.ru            }
{******************************************}

unit frxExportPDF;

interface

{$I frx.inc}

uses
{$IFNDEF Linux}
  Windows,
{$ELSE}
  LCLType, LCLIntf, LCLProc,
{$ENDIF}
  SysUtils, Graphics, Classes,
{$IFNDEF Linux}  ComObj, {$ENDIF}
  Printers, Variants, Contnrs, frxPictureGraphics,
{$IFNDEF RAD_ED}
  frxPDFSignature,
{$ENDIF}
  frxExportBaseDialog, frxClass, frxExportPDFHelpers, frxVectorCanvas, frxHelpers
{$IFDEF Delphi10}, WideStrings{$ENDIF}
{$IFDEF Delphi12}, AnsiStrings{$ENDIF}
{$IFDEF DELPHI16}, System.UITypes{$ENDIF}
{$IFDEF DEBUG_WITH_FASTMM}, FastMMDebugSupport, FastMMUsageTracker{$ENDIF}
    ;
{$IFNDEF Linux}
{$IFDEF CPUX64}
(*$HPPEMIT '#pragma link "usp10.a"'*)
{$ELSE}
(*$HPPEMIT '#pragma link "usp10.lib"'*)
{$ENDIF}
{$ENDIF}

type
  TfrxPDFEncBit = (ePrint, eModify, eCopy, eAnnot);
  TfrxPDFEncBits = set of TfrxPDFEncBit;

  TfrxPDFPage = class
  private
    FBackPictureVisible: Boolean;
    FBackPictureStretched: Boolean;
    FHeight: Double;
  public
    constructor Create(Page: TfrxReportPage);

    property Height: Double read FHeight;
    property BackPictureVisible: Boolean read FBackPictureVisible;
    property BackPictureStretched: Boolean read FBackPictureStretched;
  end;

  TfrxPDFAnnot = class
  public
    Number: Integer;
    Rect: String;
    Hyperlink: String;
    DestPage: Integer;
    DestY: Integer;
  end;

  TfrxPDFExport = class;

  TfrxPDFEngineState = class
  private
    FExport: TfrxPDFExport;
    FHeight: Extended;
    FMarginTop: Extended;
    FMarginLeft: Extended;
  public
    constructor Create(AExport: TfrxPDFExport);
    destructor Destroy; override;
    procedure BeginBBoxMode(Height: Extended);
    procedure RestoreState;
  end;

  TEmbeddedRelation = (
    erData, // The embedded file contains data which is used for the visual representation.
    erSource, // The embedded file contains the source data for the visual representation derived therefrom in the PDF part.
    erAlternative, // This data relationship should be used if the embedded data are an alternative representation of the PDF contents.
    erSupplement, // This data relationship is used if the embedded file serves neither as the source nor as the alternative representation, but the file contains additional information.
    erUnspecified // If none of the data relationships above apply or there is an unknown data relationship, this data relationship is used.
  );

  TZUGFeRD_ConformanceLevel = (
    clMINIMUM, // Minimum level.
    clBASIC,   // Basic level.
    clCOMFORT, // Comfort level.
    clEXTENDED // Extended level.
  );

  TSignatureKindSet = set of TPDFSignatureKind;
  TSignatureErrorHandling = (seShowDialog, seCancelExport, seIgnoreCertificate);
  TSignatureData = record
    Name: TComponentName;
    Description: WideString;
    Kind: TPDFSignatureKind;
    Location: WideString;
    Reason: WideString;
    ContactInfo: WideString;
    CertificatePath: WideString;
    CertificatePassword: AnsiString;
    Used: Boolean;
  end;

  TSignatureInfo = class (TPersistent)
  private
    FData: TSignatureData;
  protected
    procedure AssignTo(Dest: TPersistent); override;
  public
    constructor Create(DigitalSignatureView: TfrxDigitalSignatureView);
    constructor CreateUnknown(PDFExport: TfrxPDFExport; AName: TComponentName);
    constructor CreateData(SD: TSignatureData);

    property Data: TSignatureData read FData write FData;
  end;

  TSignatureInfoList = class (TOwnObjList)
  private
    function GetData(Index: Integer): TSignatureData;
    procedure SetData(Index: Integer; const Value: TSignatureData);
  protected
    FPDFExport: TfrxPDFExport;
  public
    constructor Create(APDFExport: TfrxPDFExport);
    procedure Init;
    function IsFind(Name: TComponentName; out Index: Integer): Boolean;
    function IsContain(Name: TComponentName): Boolean;
    procedure AddDefault;
    procedure AddData(SD: TSignatureData);
    procedure GetOldDigitalSignDataFromExport;
    procedure SetOldDigitalSignDataToExport;

    property Data[Index: Integer]: TSignatureData read GetData write SetData;
    property PDFExport: TfrxPDFExport write FPDFExport;
  end;

{$IFDEF DELPHI16}
  [ComponentPlatformsAttribute(pidWin32 or pidWin64)]
{$ENDIF}
  TfrxPDFExport = class(TfrxBaseDialogExportFilter)
  private
    FCompressed: Boolean;
    FEmbeddedFonts: Boolean;
    FEmbedProt: Boolean; { TODO : Not used. Should be removed. }
    FPrintOpt: Boolean;
    FPages: TList;
    FOutline: Boolean;
    FQuality: Integer;
    FPreviewOutline: TfrxCustomOutline;
    FSubject: WideString;
    FAuthor: WideString;
    FBackground: Boolean;
    FCreator: WideString;
    FTags: Boolean; { TODO : Not used. Should be removed. }
    FProtection: Boolean;
    FUserPassword: AnsiString;
    FOwnerPassword: AnsiString;
    FProtectionFlags: TfrxPDFEncBits;
    FKeywords: WideString;
    FTitle: WideString;
    FProducer: WideString;
    FPrintScaling: Boolean;
    FFitWindow: Boolean;
    FHideMenubar: Boolean;
    FCenterWindow: Boolean;
    FHideWindowUI: Boolean;
    FHideToolbar: Boolean;
    FTransparency: Boolean;
    FSaveOriginalImages: Boolean;

    pdf: TStream;

    FStartXRef: LongInt;
    FSignatureIndex: Integer;
    FPageTreeNo: Integer;

    FPOH: TPDFObjectsHelper;

    FWidth: Extended;
    FHeight: Extended;
    FMarginLeft: Extended;
    FMarginTop: Extended;

    FEncKey: AnsiString;
    FOPass: AnsiString;
    FUPass: AnsiString;

    FEncBits: Cardinal;
    FFileID: AnsiString;

    FLastColor: TColor;
    FLastColorResult: String;

    OutStream: TMemoryStream;

    { When an anchor is being added, two changes are made:

      - a link object is written to the document
      - a reference to the link object is added to /Annots field of the page

      FPageAnnots contains text of /Annots field.
      This stream is updated by WriteLink and its auxiliary routines. }
    FPageAnnots: TReferenceArray;
    FPageAnnotsArray: TPageByPageReferenceArray;
    FAnnots: TList;
    FXObjects: TfrxPDFXObjectArray;
    FPageXObjects: TReferenceArray;
    FPageXObjectsArray: TPageByPageReferenceArray;
    FFonts: TPDFGLobalFonts;

    FPageFonts: TReferenceArray;
    FPageFontsArray: TPageByPageReferenceArray;

    FPagesRef: TReferenceArray;
    FAcroFormsRefs: TReferenceArray;

    FMetaFileId: LongInt;
    FStructId: LongInt;
    FColorProfileId: LongInt;
    FAttachmentsNamesId: LongInt;
    FAttachmentsListId: LongInt;
    {$IFNDEF FPC}
    FPDFviaEMF: Integer;
    {$ENDIF}
    FPdfA: Boolean;
    FPDFStandard: TPDFStandard;
    FPDFVersion: TPDFVersion;
    FUsePNGAlpha: Boolean;
    FPictureDPI: Integer;
    FInteractiveForms: Boolean;

    FPDFState: TfrxPDFEngineState;
    FInteractiveFormsFontSubset: WideString;
    FGraphicHelper: TfrxCustomGraphicFormatClass;
    FAlphaGraphicHelper: TfrxCustomGraphicFormatClass;

    // Private for signature purposes
    FContentPosition, FContentEndPosition, FByteRangeIndex: Int64;
{$IFNDEF RAD_ED}
    FSignature: TfrxPDFSignature;
{$ENDIF}
    FDigitalSignLocation: WideString;
    FDigitalSignReason: WideString;
    FDigitalSignContactInfo: WideString;
    FDigitalSignCertificatePath: WideString;
    FDigitalSignCertificatePassword: AnsiString;
    FSignErrorHandling: TSignatureErrorHandling;

    FIncrementalExport: TfrxPDFExport;
    FFirstNextId: Integer;
    FSignatureData: TSignatureData;
    FCurvePrecision: Integer;

    function AddSignatureDict: Integer;
    procedure AddSignatureAppearence(sigObjNo: LongInt);
    procedure AddSignature;
    procedure AddAcroForm;

    function GetPDFDash(const LineStyle: TfrxFrameStyle; Width: Extended): String;
    function GetID: AnsiString;
    function CryptStr(Source: AnsiString; id: Integer; IsEscapeSpecialChar: Boolean = True): AnsiString;
    function CryptToHex(Source: AnsiString; id: Integer): AnsiString;

    function PrepareStr(const Text: WideString; Id: Integer): AnsiString;
    function PrepareAnsiStr(const Text: AnsiString; Id: Integer): AnsiString;
    function PrepareCreationDate(Id: Integer): AnsiString;

    function EscapeSpecialChar(TextStr: AnsiString): AnsiString;
    function StrToUTF16(const Value: WideString): AnsiString;
    function StrToUTF16H(const Value: WideString): AnsiString;
    function PMD52Str(p: Pointer): AnsiString;
    function PadPassword(Password: AnsiString): AnsiString;
    procedure PrepareKeys;
    procedure SetProtectionFlags(const Value: TfrxPDFEncBits);
    procedure Clear;
    procedure WriteFont(pdfFont: TfrxPDFFont);
    procedure AddObject(const Obj: TfrxView);
    function AddMemo(const Memo: TfrxCustomMemoView; IsInteractive: Boolean = False; SelectedLine: Integer = -1): Extended;
    procedure AddMemoField(const Memo: TfrxCustomMemoView; ViaEMF: Boolean);
    procedure AddCheckBox(const Obj: TfrxView; IsInteractiveCB: Boolean = False);
    procedure AddCheckBoxField(const Obj: TfrxView; ViaEMF: Boolean);
    procedure AddComboBox(const Obj: TfrxView; IsInteractive: Boolean = False);
    procedure AddListBox(const Obj: TfrxView; IsInteractive: Boolean = False);
    procedure AddListControlField(const Obj: TfrxView);
    procedure AddPictureField(const Obj: TfrxView);
    procedure AddLine(const Line: TfrxCustomLineView);
    procedure AddShape(const Shape: TfrxShapeView);
    procedure AddDigitalSignature(const Obj: TfrxDigitalSignatureView);
    procedure AddViaEMF(const Obj: TfrxView; IsInBBOX: Boolean = False);
    function AddAsPicture(const Obj: TfrxView): Integer;
    procedure CreateAlphaFromColorMask(TransparentColorMask: TColor; SourceGraphic: TBitmap; var MaskBytes: TMaskArray);
    procedure CreateAlphaMask(GHelper: TfrxCustomGraphicFormatClass; SourceGraphic: TBitmap; var MaskBytes: TMaskArray);

    function GetPDFColor(const Color: TColor): String;

    procedure AddAttachments;
    procedure AddEmbeddedFileItem(EmbeddedFile: TObject);
    procedure AddStructure;
    procedure AddMetaData;
    procedure AddColorProfile;
    procedure WritePDFStream(Target, Source: TStream;
      id: LongInt; Compressed, Encrypted: Boolean;
      startingBrackets, endingBrackets, enableLength2: Boolean);
    procedure SetEmbeddedFonts(const Value: Boolean);

    function GetRect(Obj: TfrxView): TfrxRect;
    function GetClipRect(Obj: TfrxView; Internal: Boolean = False): TfrxRect;
    function GetDMPRect(R: TfrxRect): TfrxRect;
    {$IFNDEF RAD_ED}
    {$IFNDEF FPC}
    function GetRectEMFExport(Obj: TfrxView): TfrxRect;
    {$ENDIF}
    {$ENDIF}
    procedure Cmd(const Args: string);
    procedure Cmd_ObjPath(Obj: TfrxView);
    procedure Cmd_RoundRectanglePath(RoundedRect: TfrxShapeView);
    procedure Cmd_EllipsePath(Ellipse: TfrxShapeView);
    procedure Cmd_TrianglePath(Triangle: TfrxShapeView);
    procedure Cmd_DiamondPath(Diamond: TfrxShapeView);
    procedure Cmd_ClipObj(Obj: TfrxView);
    procedure Cmd_FillObj(Obj: TfrxView; Color: TColor);
    procedure Cmd_FillBrush(Obj: TfrxView; BrushFill: TfrxBrushFill);
    procedure Cmd_FillGlass(Obj: TfrxView; GlassFill: TfrxGlassFill);
    procedure Cmd_FillGradient(Obj: TfrxView; GradientFill: TfrxGradientFill);
    procedure Cmd_Hatch(Obj: TfrxView; Color: TColor; Style: TBrushStyle);
    procedure Cmd_ClipRect(Obj: TfrxView);
    function Cmd_Font(Obj: TfrxView): TfrxPDFFont; overload;
    function Cmd_Font(Font: TFont): TfrxPDFFont; overload;

    procedure CmdMoveTo(x, y: Extended);
    procedure CmdLineTo(x, y: Extended);
    procedure CmdCurveTo(x1, y1, x2, y2, x3, y3: Extended);
    procedure CmdFillColor(Color: TColor);
    procedure CmdStrokeColor(Color: TColor);
    procedure CmdStroke;
    procedure CmdLineWidth(Value: Extended);

    function IsInteractiveField(Obj: TfrxView): Boolean;

    function pdfX(x: Extended): Extended;
    function pdfY(y: Extended): Extended;
    function pdfSize(Size: Extended): Extended;
    function pdfPoint(x, y: Extended): TfrxPoint;
    procedure StartBBoxMode(const Obj: TfrxView);
    function EndBBoxMode: String;
    procedure SetPDFStandard(const Value: TPDFStandard);
    procedure SetPDFVersion(const Value: TPDFVersion);
    procedure SetTransparency(const Value: Boolean);
    procedure SetPdfA(const Value: Boolean);
    procedure SetPictureDPI(const Value: Integer);
    procedure SetSaveOriginalImages(const Value: Boolean);
    procedure SetInteractiveForms(const Value: Boolean);
    procedure SetQuality(const Value: Integer);
    procedure SetCurvePrecision(const Value: Integer);

    procedure SetUserPassword(vUserPassword: AnsiString);
    procedure SetOwnerPassword(vOwnerPassword: AnsiString);
  protected
    FPageRect: TfrxRect;
    stLeft, stRight, stTop, stBottom, stRect: String;
    FEmbeddedFiles: TObjectList;
    FZUGFeRDDescription: String;
    FCreationDateTime: String;
    FCreationDateTimeMeta: String;
    FDebugDateTimeID: Boolean; // for debugging

    FSignRect: AnsiString;
    FSignatureExists: Boolean;
    FSignatureInfoList: TSignatureInfoList;
    FCurrentPageNo: Integer;
    FCurrentPageIndex: Integer;
    FSignaturePageIndex: Integer;
    FRoot: TfrxPDFExport;
    FParent: TfrxPDFExport;
    FInformationDictionaryNo: Integer;
    FInteractiveFormDictionaryNo: Integer;
    FCatalogDictionaryNo: Integer;
    FEncryptNo: LongInt;

    FDateTime: TDateTime;
    FID: AnsiString;

    function IsAddViaEMF(const Obj: TfrxView): Boolean;
    procedure AddAsPictureOld(const Obj: TfrxView);
    procedure DoFill(const Obj: TfrxView);
    procedure DoFrame(const aFrame: TfrxFrame; const aRect: TfrxRect);

    function STpdfPoint(x, y: Extended): String;
    function STpdfSize(Size: Extended): String;
    function STpdfRect(x, y, Width, Height: Extended): String;

    procedure ExportViaVector(const Memo: TfrxCustomMemoView);
    procedure Vector_ExtTextOut(Memo: TfrxCustomMemoView; Vector: TVector_ExtTextOut);

    procedure WritePageTree;
    procedure WriteInformationDictionary;
    procedure WriteCatalogDictionary(OutlineObjNo: integer);
    procedure OutUsedXObjects;

    function IsRoot: Boolean;

    property Root: TfrxPDFExport read FRoot;
    property Parent: TfrxPDFExport read FParent;
  public
    property SignatureInfoList: TSignatureInfoList read FSignatureInfoList;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    class function GetDescription: String; override;
    class function ExportDialogClass: TfrxBaseExportDialogClass; override;
//    function ShowModal: TModalResult; override;
    function Start: Boolean; override;
    procedure ExportObject(Obj: TfrxComponent); override;
    procedure Finish; override;
    procedure ExecuteIncremental;
    procedure StartPage(Page: TfrxReportPage; Index: Integer); override;
    procedure FinishPage(Page: TfrxReportPage; Index: Integer); override;
    procedure BeginClip(Obj: TfrxView); override;
    procedure EndClip; override;
    procedure AddEmbeddedFile(Name, Description: String; ModDate: TDateTime;
      Relation: TEmbeddedRelation; MIME: String; FileStream: TStream);
    procedure AddEmbeddedXML(Name, Description: String; ModDate: TDateTime;
      FileStream: TStream; ZUGFeRDLevel: TZUGFeRD_ConformanceLevel = clBASIC; const ZUGFeRDDescription: string = '');
    function IsPDFA: Boolean;
    function IsPDFA_1: Boolean;
    function IsSignatureExists(SignatureKindSet: TSignatureKindSet = [Low(TPDFSignatureKind) .. High(TPDFSignatureKind)]): Boolean;

    procedure FillSignatureInfoList(SIL: TSignatureInfoList);

    property SaveOriginalImages: Boolean read FSaveOriginalImages write SetSaveOriginalImages;
    property PictureDPI: Integer read FPictureDPI write SetPictureDPI;
    /// <summary>
    ///   Relevant for a large number of curves, like, when using TfrxPDFView.
    ///   2 - high quality + big size (default)
    ///   1 - medium quality + medium size
    ///   0 - low quality + small size
    /// </summary>
    property CurvePrecision: Integer read FCurvePrecision write SetCurvePrecision;
    property UsePNGAlpha: Boolean read FUsePNGAlpha write FUsePNGAlpha;
    property DigitalSignLocation: WideString read FDigitalSignLocation write FDigitalSignLocation;
    property DigitalSignReason: WideString read FDigitalSignReason write FDigitalSignReason;
    property DigitalSignContactInfo: WideString read FDigitalSignContactInfo write FDigitalSignContactInfo;
    property DigitalSignCertificatePath: WideString read FDigitalSignCertificatePath write FDigitalSignCertificatePath;
    property DigitalSignCertificatePassword: AnsiString read FDigitalSignCertificatePassword write FDigitalSignCertificatePassword;
    property SignErrorHandling: TSignatureErrorHandling read FSignErrorHandling write FSignErrorHandling default seShowDialog;
    property IncrementalExport: TfrxPDFExport read FIncrementalExport;
  published
    property Compressed: Boolean read FCompressed write FCompressed default True;
    property EmbeddedFonts: Boolean read FEmbeddedFonts write SetEmbeddedFonts default False;
    property EmbedFontsIfProtected: Boolean read FEmbedProt write FEmbedProt default True;  { TODO : Not used. Should be removed. }
    property InteractiveForms: Boolean read FInteractiveForms write SetInteractiveForms default False;
    property InteractiveFormsFontSubset: WideString read FInteractiveFormsFontSubset write FInteractiveFormsFontSubset;
    property OpenAfterExport;
    property PrintOptimized: Boolean read FPrintOpt write FPrintOpt;
    property Outline: Boolean read FOutline write FOutline;
    property Background: Boolean read FBackground write FBackground;
    property HTMLTags: Boolean read FTags write FTags; { TODO : Not used. Should be removed. }
    property OverwritePrompt;
    property Quality: Integer read FQuality write SetQuality;
    property Transparency: Boolean read FTransparency write SetTransparency default True;

    property Title: WideString read FTitle write FTitle;
    property Author: WideString read FAuthor write FAuthor;
    property Subject: WideString read FSubject write FSubject;
    property Keywords: WideString read FKeywords write FKeywords;
    property Creator: WideString read FCreator write FCreator;
    property Producer: WideString read FProducer write FProducer;

    property UserPassword: AnsiString read FUserPassword write SetUserPassword;
    property OwnerPassword: AnsiString read FOwnerPassword write SetOwnerPassword;
    property ProtectionFlags: TfrxPDFEncBits read FProtectionFlags write SetProtectionFlags;

    property HideToolbar: Boolean read FHideToolbar write FHideToolbar;
    property HideMenubar: Boolean read FHideMenubar write FHideMenubar;
    property HideWindowUI: Boolean read FHideWindowUI write FHideWindowUI;
    property FitWindow: Boolean read FFitWindow write FFitWindow;
    property CenterWindow: Boolean read FCenterWindow write FCenterWindow;
    property PrintScaling: Boolean read FPrintScaling write FPrintScaling;

    property PdfA: Boolean read FPdfA write SetPdfA; // Deprecated
    property PDFStandard: TPDFStandard read FPDFStandard write SetPDFStandard;
    property PDFVersion: TPDFVersion read FPDFVersion write SetPDFVersion;
  end;

  { Returns a color in PDF form. }

function PdfColor(Color: TColor): AnsiString;

{ Returns a pair of coordinates in PDF form. }

function pdfPoint(x, y: Double): AnsiString;

{ Moves the pen to the specified point. }

function PdfMove(x, y: Double): AnsiString;

{ Draws a line to the specified point. }

function PdfLine(x, y: Double): AnsiString;

{ Changes the current color. }

function PdfSetColor(Color: TColor): AnsiString;

{ Changes width of the line drawed by the pen.
  The width is measured in points (1/72 of an inch). }

function PdfSetLineWidth(Width: Double): AnsiString;

{ Changes the color of the pen. }

function PdfSetLineColor(Color: TColor): AnsiString;

{ Fills the latest contoured area. }

function PdfFill: AnsiString;

{ Strokes the latest contoured area. }

function PdfStroke: AnsiString;

{ Fills a rectangle area. }

function PdfFillRect(R: TfrxRect; Color: TColor): AnsiString;

{ Strokes a rectangle area. }

function PdfStrokeRect(R: TfrxRect; Color: TColor; LineWidth: Extended)
  : AnsiString;

{ Returns either (...) or <...> sequence. }

function PdfString(const Str: WideString): AnsiString;

const
  DefaultSignatureIndex = 0;

implementation
end.

