
{******************************************}
{                                          }
{             FastReport VCL               }
{            PDF export filter             }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxExportPDFHelpers;

interface

{$I frx.inc}

uses
{$IFNDEF Linux}
  Windows,
{$ELSE}
  LCLType, LCLIntf, LCLProc,
{$ENDIF}
  Graphics, Classes, SysUtils, frxRC4, frxTrueTypeCollection, Contnrs, frxBaseGraphicsTypes, frxPictureGraphics,
{$IFNDEF FPC}frxEMFAbstractExport, {$ENDIF}
  frxClass,
{$IFNDEF FPC}frxEMFFormat, {$ENDIF}
{$IFDEF FPC}LazHelper, {$ENDIF}
  frxUtils,
{$IFDEF Delphi12}
  AnsiStrings,
{$ENDIF}
{$IFDEF DELPHI16}
  System.UITypes,
{$ENDIF}
  frxAnaliticGeometry, frxExportHelpers, frxHelpers
{$IFDEF Linux}
  , types,  Messages, Controls, Forms, Dialogs, Process,
  Variants,  StdCtrls, Db, DBCtrls, Buttons, ExtCtrls, Menus, ComCtrls, lmessages, frxStorage, frxLinuxFonts
{$ENDIF};

type
  TfrxPDFFont = class(TfrxExportFont)
  private
    FColor: TColor;
    FFontName: AnsiString;
    FSize: Extended;
  protected
//    Index: Integer;
    FSaved: Boolean;

    procedure FillOutlineTextMetrix();
  public
    Name: AnsiString;
    Reference: Longint;

    constructor Create(Font: TFont);
    procedure Save(AReference: Longint);
    procedure PackTTFFont;
    function GetFontName: AnsiString;
    function SpaceAdjustment(RS: TRemapedString; TextWidth, FontSize: Extended): Extended;
    function FontHeightToPointSizeFactor: Double;
    {$IFDEF Linux}
    function GetFontDataSize(): Longint; Override;
    {$ENDIF}
    procedure Update(const SourcePDFFont: TfrxPDFFont; const SourceFont: TFont);
    property FontName: AnsiString read FFontName write FFontName;
    property Size: Extended read FSize write FSize;
    property Color: TColor read FColor write FColor;
    property Saved: Boolean read FSaved;
  end;

  TfrxPDFXObjectHash = array[0..15] of Byte; // MD5
  TfrxPDFXObjectArray = class
  private
    FCount: Integer;
    FObjId: array of Integer; // id that appears in 'id 0 R'
    FHash: array of TfrxPDFXObjectHash;

    function GetId(Index: Integer): Integer;
  protected
    procedure SureIndex(Index: Integer);
  public
    constructor Create;
    function Add(AId: Integer; const Hash: TfrxPDFXObjectHash): Integer;
    procedure Clear;
    function Find(const Hash: TfrxPDFXObjectHash): Integer;

    property Id[Index: Integer]: Integer read GetId;
    property Count: Integer read FCount;
  end;

  TCrossReferenceSubsection = record
    Start: Boolean;
    Length: Integer; // number of entries
  end;

  TReferenceArray = class
  private
    FCount: Integer;
    FData: array of Integer;
    function GetRef(Index: Integer): string;
    function GetNo(Index: Integer): Integer;
  public
    constructor Create;
    procedure Add(ObjNo: Integer);
    procedure WriteToStream(Stream: TStream; Caption: string);
    procedure Clear;

    property Count: Integer read FCount;
    property Ref[Index: Integer]: string read GetRef;
    property No[Index: Integer]: Integer read GetNo;
  end;

  TPageByPageReferenceArray = class
  private
    FData: array of TReferenceArray;
    FContents: array of Integer;
    function GetData(Index: Integer): TReferenceArray;
    procedure SetData(Index: Integer; const Value: TReferenceArray);
    function GetContents(Index: Integer): Integer;
    procedure SetContents(Index: Integer; const Value: Integer);
  protected
    procedure SureIndex(Index: Integer);
  public
    constructor Create;
    destructor Destroy; override;

    property Data[Index: Integer]: TReferenceArray read GetData write SetData; default;
    property Contents[Index: Integer]: Integer read GetContents write SetContents;
  end;

  TCrossReferenceSection = class (TOwnObjList)
  private
    FSingleCharacterEOL: Boolean;
    FSectionOffset: Integer;
    FNextId: Integer;
  protected
    function IndexOfId(Id: Integer): Integer;
  public
    constructor Create(ASingleCharacterEOL: Boolean = False);
    procedure Clear; override;
    procedure Init;
    procedure AddId(Id: Integer; Offset: Integer = 0);
    function AddNext(Offset: Integer = 0): Integer; // Return Id
    procedure SetOffset(Id: Integer; Offset: Integer);

    procedure PrepareToOut;
    function IsSubsection(Index: Integer; out SubsectionStr: string): Boolean;
    function OutLine(Index: Integer): string;

    property NextId: Integer read FNextId write FNextId;
    property SectionOffset: Integer write FSectionOffset;
  end;

  TPDFGLobalFonts = class(TOwnObjList)
  protected
    function Get(Index: Integer): TfrxPDFFont;
  public
    function AddFont(const Font: TFont; const OSFontName: String): Integer;
    function IndexOfFont(Name: String; Style: TFontStyles): Integer;
    function IndexByReference(No: Integer): Integer;

    property Items[Index: Integer]: TfrxPDFFont read Get; default;
  end;

  TPDFObjectsHelper = class
  private
    FFonts: TPDFGLobalFonts;
    FAcroFonts: TList;
    FBBoxFonts: TList;
    FPageFonts: TReferenceArray; // Fonts' ids used within the current page
    FCRS: TCrossReferenceSection;
    FpdfStream: TStream;
    FProtection: boolean;
    FEncKey: AnsiString;
    FEmbedded: boolean;
    FQuality: Integer;
    FXObjects: TfrxPDFXObjectArray;
    FPageXObjects: TReferenceArray; // XObjects' ids used within the current page
    FIsBBoxReleative: Boolean;
    FLastRequestedXRef: Integer;
    FGraphicHelper: TfrxCustomGraphicFormatClass;
    function GetAcroFonts(Index: integer): TfrxPDFFont;
    function GetAcroFontsCount: integer;
    function GetBBoxFonts(Index: integer): TfrxPDFFont;
    function GetBBoxFontsCount: Integer;
    function GetLastObjectXRefID: Integer;
  protected
    function GetOSFontName(const Font: TFont): String;
    function IsFontNameAndStyle(const Font: TFont; Name: String; Style: TFontStyles): Boolean;
    function GetGlobalFont(const Font: TFont; const OSFontName: String): TfrxPDFFont;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    function GetAcroFont(const Font: TFont): TfrxPDFFont;
    function GetObjFontNumber(const Font: TFont): integer; // Returns the index in the FFonts, if absent, then first adds

    procedure OutPageFonts;
    function FindXObject(const Hash: TfrxPDFXObjectHash): Integer;

    function OutXObjectImage(XObjectHash: TfrxPDFXObjectHash;
      Image: TGraphic; XObjectStream: TStream;
      IsTransparent: Boolean = False; MaskId: Integer = 0): Integer;
    function OutTransparentGraphic(Graphic: TGraphic): Integer; // map images
    function UpdateXRef: Integer;
    procedure StartBBox;
    procedure EndBBox;

    property AcroFonts[Index: integer]: TfrxPDFFont read GetAcroFonts;
    property AcroFontsCount: integer read GetAcroFontsCount;
    property BBoxFonts[Index: integer]: TfrxPDFFont read GetBBoxFonts;
    property BBoxFontsCount: Integer read GetBBoxFontsCount;

    property Fonts: TPDFGLobalFonts read FFonts write FFonts;

    property XObjects: TfrxPDFXObjectArray read FXObjects write FXObjects;
    property PageXObjects: TReferenceArray read FPageXObjects write FPageXObjects;
    property PageFonts: TReferenceArray read FPageFonts write FPageFonts;
    property pdfStream: TStream write FpdfStream;
    property CRS: TCrossReferenceSection read FCRS;
    property LastObjectXRefID: Integer read GetLastObjectXRefID;
    property Protection: boolean write FProtection;
    property EncKey: AnsiString write FEncKey;
    property EmbeddedFonts: Boolean read FEmbedded write FEmbedded default False;
    property IsBBox: Boolean read FIsBBoxReleative;
    property Quality: Integer read FQuality write FQuality;
  end;

function IsNeedsItalicSimulation(Font: TFont; out Simulation: String): Boolean;
function IsNeedsBoldSimulation(Font: TFont; out Simulation: String): Boolean;

procedure AddStyleSimulation(FontName: String; FontStyles: TFontStyles);
procedure DeleteStyleSimulation(FontName: String);

function StrToHex(const Value: WideString): AnsiString;
function StrToHexSp(const Value: WideString; SpaceAdjustment: Extended): AnsiString;
function StrToHexDx(const RS: TRemapedString; pdfDX: TDoubleArray; AverageDx: Boolean): AnsiString;

function Color2Str(Color: TColor): String;
function frxRect2Str(DR: TfrxRect; const Prec: Integer = DefaultPrec): String;

function frxPointSum(P1, P2: TfrxPoint): TfrxPoint;
function frxPointMult(P: TfrxPoint; Factor: Extended): TfrxPoint;

procedure GetStreamHash(out Hash: TfrxPDFXObjectHash; S: TStream);
function ObjNumberRef(FNumber: longint): String;
function ObjNumber(FNumber: longint): String;
function PrepXRefPos(Index: Integer): String;
function CryptStream(Source: TStream; Target: TStream; Key: AnsiString; id: Integer): AnsiString;

procedure Write(Stream: TStream; const S: AnsiString);{$IFDEF Delphi12} overload;
procedure Write(Stream: TStream; const S: String); overload; {$ENDIF}
procedure WriteLn(Stream: TStream; const S: AnsiString);{$IFDEF Delphi12} overload;
procedure WriteLn(Stream: TStream; const S: String); overload; {$ENDIF}

type
  TMaskArray = array of byte;
function BitmapPixelSize(Bitmap: TBitmap): Integer;
function CreateBitmap(PixelFormat: TPixelFormat; SizeBitmap: TBitmap): TBitmap; overload;
function CreateBitmap(PixelFormat: TPixelFormat; Width, Height: Integer): TBitmap; overload;
procedure CreateMask(Bitmap: TBitmap; var Mask: TMaskArray);
procedure SaveMask(pdf, XObjectStream: TStream; MaskBytes: TMaskArray;
  FPOH: TPDFObjectsHelper; TempBitmap: TBitmap; FProtection: Boolean; FEncKey: AnsiString;
  out XObjectHash: TfrxPDFXObjectHash; out XMaskId, PicIndex: Integer);

const
  PDF_DIVIDER = 0.75;

type
  TfrxShapeKindSet = set of TfrxShapeKind;

function IsShape(Obj: TfrxView; ShapeKindSet: TfrxShapeKindSet): boolean;
function Is2DShape(Obj: TfrxView): boolean;

type
  TViewSizes = record l, t, w, h, r, b: Extended; end;

function ShadowlessSizes(Obj: TfrxView): TViewSizes;

type
  TRGBAWord = packed record
    Blue: Byte;
    Green: Byte;
    Red: Byte;
    Alpha: Byte;
  end;

  PRGBAWord = ^TRGBAWordArray;
  TRGBAWordArray = array[0..4095] of TRGBAWord;

  TPDFStandard =
    (psNone, psPDFA_1a, psPDFA_1b, psPDFA_2a, psPDFA_2b, psPDFA_3a, psPDFA_3b);
  TPDFVersion =
    (pv14, pv15, pv16, pv17);

const
  PDFStandardName: array[TPDFStandard] of string =
    ('None', 'PDF/A-1a', 'PDF/A-1b', 'PDF/A-2a', 'PDF/A-2b', 'PDF/A-3a', 'PDF/A-3b');
  PDFPartName: array[TPDFStandard] of string =
    (    '',       '1',        '1',         '2',       '2',         '3',       '3');
  PDFConformanceName: array[TPDFStandard] of string =
    (    '',       'A',        'B',         'A',       'B',         'A',       'B');
  PDFVersionName: array[TPDFVersion] of string =
    ('1.4', '1.5', '1.6', '1.7');

function PDFStandardByName(StandardName: string): TPDFStandard;
function PDFVersionByName(VersionName: string): TPDFVersion;
function IsPDFA(ps: TPDFStandard): Boolean;
function IsPDFA_1(ps: TPDFStandard): Boolean;
function IsVersionByStandard(ps: TPDFStandard; var pv: TPDFVersion): Boolean;

function PDFFontSize(Font: TFont): Extended;
function EMFPDFFontSize(Font: TFont): Extended;
//function IsMonospaced(FontName: string): Boolean;
procedure StretchFont(FontName: TFontName; FontSize: Integer; Factor: Double); overload; // Monospaced Font
procedure StretchFont(FontName: TFontName; Factor: Single); overload; // Proportionsl Font (for EMF)
procedure ClearStretchedFont(FontName: TFontName);
procedure DisableTpPtCorrection;
procedure EnableTpPtCorrection;

type
  TInt64List = class
  private
    function GetItem(Index: integer): Int64;
    procedure Setitem(Index: integer; const Value: Int64);
    function GetCount: Integer;
  protected
    FObjectList: TObjectList;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Add(Value: Int64);
    procedure Clear;

    property Items[Index: integer]: Int64 read GetItem write Setitem; default;
    property Count: Integer read GetCount;
  end;

const
  tpPt = 0.254 / 72; // typography pt
  UnderlineShift = -0.12;
  StrikeOutShift = 0.28;
  UnderlineWidth = 0.08;
  StrikeOutWidth = UnderlineWidth / 1.4142;


implementation
end.

