
{******************************************}
{                                          }
{             FastReport VCL               }
{               DB dataset                 }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxDBSet;

interface

{$I frx.inc}

uses
  SysUtils, {$IFNDEF FPC}Windows, Messages,{$ENDIF} Classes, frxClass, DB
{$IFDEF FPC}
  , LazHelper, LCLType, LazarusPackageIntf
{$ENDIF}
{$IFDEF Delphi6}
, Variants
{$ENDIF}
{$IFDEF Delphi10}
, WideStrings
{$ENDIF};


type
  TfrxDataSetOption = (dsoStringAsDisplayText);
  TfrxDataSetOptions = set of TfrxDataSetOption;
{$IFDEF DELPHI16}
/// <summary>
///   The TfrxDBDataSet component is designed for connecting to the DB
///   components, which are based on TDataSet, such as "TTable" and "TQuery".
///   Use DataSource or DataSet properties to connect to the data.
/// </summary>
[ComponentPlatformsAttribute(pidWin32 or pidWin64)]
{$ENDIF}
  TfrxDBDataset = class(TfrxCustomDBDataset)
  private
    FBookmark: TBookmark;
    FDataSet: TDataSet;
    FDataSource: TDataSource;
    FDS: TDataSet;
    FEof: Boolean;
    FBCDToCurrency: Boolean;
    FSaveOpenEvent: TDatasetNotifyEvent;
    FSaveCloseEvent: TDatasetNotifyEvent;
    FSaveAfterRefresh: TDatasetNotifyEvent;
    FSaveFiedsChange: TNotifyEvent;
    FNeedFieldsUpdate: Boolean;
    FDSOptions: TfrxDataSetOptions;
    procedure BeforeClose(Sender: TDataSet);
    procedure AfterOpenEvent(Sender: TDataset);
    procedure AfterRefresh(Sender: TDataset);
    procedure FieldsChanged(Sender: TObject);
    procedure SetDataSet(Value: TDataSet);
    procedure SetDataSource(Value: TDataSource);
    function DataSetActive: Boolean;
    function IsDataSetStored: Boolean;
    procedure UpdateFields;
  protected
    function GetDisplayText(Index: String): WideString; override;
    function GetDisplayWidth(Index: String): Integer; override;
    function GetFieldType(Index: String): TfrxFieldType; override;
    function GetValue(Index: String): Variant; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    {$IFDEF FPC}
    TempTag:Integer;
    {$ENDIF}
    procedure Initialize; override;
    procedure Finalize; override;
    procedure First; override;
    procedure Next; override;
    procedure Prior; override;
    procedure Open; override;
    procedure Close; override;
    function Eof: Boolean; override;

    function GetDataSet: TDataSet;
    function IsBlobField(const fName: String): Boolean; override;
    function IsWideMemoBlobField(const fName: String): Boolean; override;
    function IsMemoBlobField(const fName: String): Boolean; override;
    function IsHasMaster: Boolean; override;
    function RecordCount: Integer; override;
    procedure AssignBlobTo(const fName: String; Obj: TObject); override;
    procedure GetFieldList(List: TStrings); override;
  published
    /// <summary>
    ///   The data set - any TDataSet descendant.
    /// </summary>
    property DataSet: TDataSet read FDataSet write SetDataSet stored IsDataSetStored;
    /// <summary>
    ///   The data source - TDataSource -&gt; TDataSet.
    /// </summary>
    property DataSource: TDataSource read FDataSource write SetDataSource stored IsDataSetStored;
    /// <summary>
    ///   Converts fields of type BCD to Currency if set to True.
    /// </summary>
    property BCDToCurrency: Boolean read FBCDToCurrency write FBCDToCurrency;
    property DataSetOptions: TfrxDataSetOptions read FDSOptions write FDSOptions;
  end;

{$IFDEF FPC}
//  procedure Register;
{$ENDIF}
implementation
end.
