
{******************************************}
{                                          }
{             FastReport VCL               }
{          Data Tree tool window           }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxDataTree;

interface

{$I frx.inc}

uses
  {$IFNDEF FPC}
  Windows, Messages,
  {$ENDIF}
  SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, frxClass, fs_xml, ComCtrls, frxBaseForm, frxComCtrls, Menus, frxDock
{$IFDEF UseTabset}
, Tabs
{$ENDIF}
{$IFDEF Delphi6}
, Variants
{$ENDIF}
{$IFDEF DELPHI16}
, System.Types
{$ENDIF};


type
  TfrxDataTreeActions = (dtaNode = 0, dtaDataEdit = 53, dtaVariablesEdit = 52, dtaSortData = 60, dtaUnsorted = 127, dtaAscending = 125, dtaDescending = 126, dtaCollapse = 117, dtaExpand = 118, dtadtfInsCaption = 123, dtaInsField = 124);
  TfrxDataTreeFlag = (dtfInsField, dtfInsCaption);

  TfrxDataTreeFlags = set of TfrxDataTreeFlag;

  TfrxDataTreeForm = class(TfrxDockForm)
    FunctionsPn: TPanel;
    Splitter1: TSplitter;
    HintPanel: TScrollBox;
    FunctionDescL: TLabel;
    FunctionNameL: TLabel;
    NoDataPn: TScrollBox;
    NoDataL: TLabel;
    procedure FormResize(Sender: TObject);
    procedure DataTreeCustomDrawItem(Sender: TCustomTreeView;
      Node: TTreeNode; State: TCustomDrawState; var DefaultDraw: Boolean);
    procedure FunctionsTreeChange(Sender: TObject; Node: TTreeNode);
    procedure DataTreeDblClick(Sender: TObject);
    procedure ClassesTreeExpanding(Sender: TObject; Node: TTreeNode;
      var AllowExpansion: Boolean);
    procedure ClassesTreeCustomDrawItem(Sender: TCustomTreeView;
      Node: TTreeNode; State: TCustomDrawState; var DefaultDraw: Boolean);
    procedure SortCBClick(Sender: TObject);
  private
    { Private declarations }
    FXML: TfsXMLDocument;
    FImages: TImageList;
    FReport: TfrxReport;
    FUpdating: Boolean;
    FFirstTime: Boolean;
    FMultiSelectAllowed: Boolean;
{$IFDEF UseTabset}
    FTabs: TTabSet;
{$ELSE}
    FTabs: TTabControl;
{$ENDIF}
    FDataTree: TfrxTreeView;
    FClassesTree: TfrxTreeView;
    FDataPanel: TfrxTreePanel;
    FClassesPanel: TfrxTreePanel;
    FVariablesPanel: TfrxTreePanel;
    FVariablesTree: TfrxTreeView;
    FFunctionsPanel: TfrxTreePanel;
    FFunctionsTree: TfrxTreeView;
    FDataTreeFlags: TfrxDataTreeFlags;
    FDataTreeSortType: TfrxTreeSortType;
    FSortPopUp: TPopupMenu;
    FSortButton: TfrxToolPanelButton;
    FVariablesButton: TfrxToolPanelButton;
    FActionBtnList: array[TfrxDataTreeFlag] of TfrxToolPanelButton;
    FUpdateLocked: Boolean;
    procedure FillClassesTree;
    procedure FillDataTree;
    procedure FillFunctionsTree;
    procedure FillVariablesTree;
    procedure CollapseExpand(aExpand: Boolean = False);
    procedure TabsChange(Sender: TObject);
    function GetCollapsedNodes: String;
    procedure ToolOnClick(Sender: TObject);
    function FilterCompare(Sender: TObject; Node: TTreeNode; const aFilter: string): Boolean;
    procedure CreateSortPopup;
    procedure CreatefrxTreePanel(var TreePanel: TfrxTreePanel; var aTreeView: TfrxTreeView);
    procedure CreateDataTreeView;
    procedure CreateClassesTreeView;
    procedure CreateVariablesTreeView;
    procedure CreateFunctionsTreeView;
    function GetClassesTree: TTreeView;
    function GetDataTree: TTreeView;
    function GetFunctionsTree: TTreeView;
    function GetVariablesTree: TTreeView;
    procedure SetDataTreeFlags(const Value: TfrxDataTreeFlags);
    procedure SetDataTreeSortType(const Value: TfrxTreeSortType);
    procedure SetReport(const Value: TfrxReport);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetColor_(AColor: TColor);
    procedure SetControlsParent(AParent: TWinControl);
    procedure SetLastPosition(p: TPoint);
    procedure ShowTab(Index: Integer);
    procedure UpdateItems;
    procedure UpdateSelection;
    procedure UpdateSize;
    procedure UpdateFormPPI(aNewPPI: Integer); override;
    procedure CheclMultiSelection;
    procedure DisableUpdate;
    procedure EnableUpdate;
    function GetActivePage: Integer;
    function GetFieldName(SelectionIndex: Integer = -1): String;
    function GetDataSet(SelectionIndex: Integer): TfrxDataSet;
    function ActiveDS: String;
    function GetLastPosition: TPoint;
    function IsDataField: Boolean;
    function GetSelectionCount: Integer;
    function IsDataTree(aSource: TObject): Boolean;
    procedure LoadFormPreferences(PreferencesStorage: TObject; DefPreferencesStorage: TObject); override;
    procedure SaveFormPreferences(PreferencesStorage: TObject; DefPreferencesStorage: TObject); override;
    property DataTreeFlags: TfrxDataTreeFlags read FDataTreeFlags write SetDataTreeFlags;
    property DataTreeSortType: TfrxTreeSortType read FDataTreeSortType write SetDataTreeSortType;
    property Report: TfrxReport read FReport write SetReport;
    property MultiSelectAllowed: Boolean read FMultiSelectAllowed write FMultiSelectAllowed;
    { back compat }
    property ClassesTree: TTreeView read GetClassesTree;
    property DataTree: TTreeView read GetDataTree;
    property FunctionsTree: TTreeView read GetFunctionsTree;
    property VariablesTree: TTreeView read GetVariablesTree;
  end;


implementation
end.
